# -*-python-*-

__package__    = "pocketMMap.py"
__version__    = "0.1"
__author__     = "Aaron Straup Cope"
__url__        = "http://www.aaronland.info/python/pocketMMap"
__date__       = "$Date: 2008/08/27 07:44:21 $"
__copyright__  = "Copyright (c) 2007-2008 Aaron Straup Cope. BSD license : http://www.modestmaps.com/license.txt"

# Hey look !! This is version 0.1 !!!
#
# This means it works but it also means
# there are lots of niggly details and 
# features to work out still. As of this
# writing they include:
#
# proper documentation (see __main__ instead)
#
# font sizes and the pagination widgets
# (automatic resizing and sanity checking)
#
# consider backporting the pagination widgets
# to not require pycairo
#
# markers and indexes
#
# ws-pocketmmap
# 
# merging and/or updating chunks to both
# Modest Maps and ws-compose for better
# code re-use

import ModestMaps

import array
import math
import sys

import Image
import ImageDraw
import cairo

from reportlab.pdfgen import canvas

class pocketMMap :

    def __init__ (self, height, width, margin=0, dpi=72) :

        self.paper_height = height
        self.paper_width = width
        self.paper_margin = margin
        self.paper_dpi = dpi

        self.height = 0
        self.width = 0
        self.margin = 0
        
        self.map_width = 0
        self.map_height = 0

        self.sect_width = 0
        self.sect_height = 0

        self.pages = 0
        self.sheets = 0

    	self.bbox = []
        self.zoom = 0

        # used to map a page number "index"
        # to a section key
        
        # this assumes somthing like:
        # 
        # a | b
        # c | d
        # e | f
        # g | h
        
        self.page_map = {
            '1':'c',
            '2':'e',
            '3':'g',
            '4':'h',
            '5':'f',
            '6':'d',
            '7':'b',
            '8':'a',
        }

        # pieces is a list of list mapping
        # rows and cols to rows
        #
        # maps a given piece/page to it's
        # row and col index
        
        self.pieces = []
        self.pieces_map = {}

        # the reverse map maps row-col keys to
        # page numbers

        self.reverse_map = {}

        # the sect map maps section keys (a,b,c)
        # to hashes of meta data for that section
        # stuff like height, width and offsets
        
        self.sect_map = {}

        # the source map generated by modest maps
        
        self.map_image = None

        # the individual pages with tiles (for the
        # source map) the laid out and labeled
        
        self.pages_rendered = []    

        # Modest Maps stuff
        
        self.mm = None
        self.mm_provider = None
        
    #
    #
    #

    def load_provider (self, label) :

        # Please for this method to be in MM itself...
        
        if label == 'MICROSOFT_ROAD':
            self.mm_provider = ModestMaps.Microsoft.RoadProvider()
        
        elif label == 'MICROSOFT_AERIAL':
            self.mm_provider = ModestMaps.Microsoft.AerialProvider()
        
        elif label == 'MICROSOFT_HYBRID':
            self.mm_provider = ModestMaps.Microsoft.HybridProvider()
        
        elif label == 'GOOGLE_ROAD':
            self.mm_provider = ModestMaps.Google.RoadProvider()
        
        elif label == 'GOOGLE_AERIAL':
            self.mm_provider = ModestMaps.Google.AerialProvider()
        
        elif label == 'GOOGLE_HYBRID':
            self.mm_provider = ModestMaps.Google.HybridProvider()
        
        elif label == 'YAHOO_ROAD':
            self.mm_provider = ModestMaps.Yahoo.RoadProvider()
        
        elif label == 'YAHOO_AERIAL':
            self.mm_provider = ModestMaps.Yahoo.AerialProvider()
        
        elif label == 'YAHOO_HYBRID':
            self.mm_provider = ModestMaps.Yahoo.HybridProvider()
        
        elif label == 'BLUE_MARBLE':
            self.mm_provider = ModestMaps.BlueMarble.Provider()
        
        elif label == 'OPEN_STREET_MAP':
            self.mm_provider = ModestMaps.OpenStreetMap.Provider()

        else :
            return None

    #
    #
    #
    
    def draw (self, bbox, zoom, out=None) :

        self.setup(bbox, zoom)
        
        self.fetch_map()
        self.slice_map()
        self.layout_map()

        if out :
            self.save(out)
        
    #
    #
    #
    
    def fetch_map (self) :

        coord, offset = ModestMaps.calculateMapExtent(self.mm_provider,
                                                      self.map_width, self.map_height,
                                                      ModestMaps.Geo.Location(self.bbox[0], self.bbox[1]),
                                                      ModestMaps.Geo.Location(self.bbox[2], self.bbox[3]))
        
        dim = ModestMaps.Core.Point(self.map_width, self.map_height)
        self.mm = ModestMaps.Map(self.mm_provider, dim, coord, offset)            
        
        self.map_image = self.mm.draw()
        self.map_image = self.map_image.convert("RGBA")
        
    #
    #
    #
    
    def slice_map (self) :

        #
        # Split the map into multiple pieces
        #
        
        offset_x = 0
        offset_y = 0
        row = 0
        col = 0
        
        i = 1

        while offset_y < self.map_height :

            cols_for_row = []
    
            while offset_x < self.map_width :

                x1 = offset_x
                y1 = offset_y
                x2 = x1 + self.sect_width
                y2 = y1 + self.sect_height
                
                box = (x1, y1, x2, y2)
                
                sect = self.map_image.crop(box)
                cols_for_row.append(sect)
        
                self.pieces_map[i] = (row, col)

                reverse_key = "%s-%s" % (row, col)
                self.reverse_map[reverse_key] = i
                
                offset_x += self.sect_width
                
                col += 1
                i += 1
        
            self.pieces.append(cols_for_row)

            row += 1
            col = 0
    
            offset_x = 0
            offset_y += self.sect_height

    #
    #
    #
    
    def layout_map(self) :

        first = 1
        last = (self.sheets * 8) + 1
        offset = 4
        
        for current_sheet in range(1, (self.sheets + 1)) :
    
            pm = Image.new('RGBA', (self.height, self.width), 'white')

            for i in range(first, (first + offset)) :
                self.draw_page(pm, i)
        
            for i in range((last - offset), last) :
                self.draw_page(pm, i)

            self.pages_rendered.append(pm)
    
            first += offset
            last -= offset
        
    #
    #
    #
    
    def save (self, out) :

        # reportlab docs say out can be any IO
        # thingy or a path
        
        # note the part where we (re) flip x, y
        # and height, width; see notes in setup()
        
        pdf = canvas.Canvas(out, pagesize=(self.height, self.width))

        for im in self.pages_rendered :    
            pdf.drawInlineImage(im, 0, 0)
            pdf.showPage()
    
        pdf.save()

    #
    #
    #
    
    def setup (self, bbox, zoom) :

	self.bbox = bbox
    	self.zoom = zoom
        
        # this piece is copied from wscompose/__init__.py    
        
        offset_lat = (self.bbox[2] - self.bbox[0]) / 2
        offset_lon = (self.bbox[3] - self.bbox[1]) / 2

        lat = self.bbox[0] + offset_lat
        lon = self.bbox[1] + offset_lon
        
        loc = ModestMaps.Geo.Location(lat, lon)
        coordinate = self.mm_provider.locationCoordinate(loc)            
        coordinate = coordinate.zoomTo(self.zoom)
        
        coord, offset = ModestMaps.calculateMapCenter(self.mm_provider, coordinate)
        dim = ModestMaps.Core.Point(0, 0)
        
        self.mm = ModestMaps.Map(self.mm_provider, dim, coord, offset)
        
        # this piece is copied from ModestMaps/__init__.py

        sw = ModestMaps.Geo.Location(self.bbox[0], self.bbox[1])
        ne = ModestMaps.Geo.Location(self.bbox[2], self.bbox[3])
        nw = ModestMaps.Geo.Location(ne.lat, sw.lon)
        se = ModestMaps.Geo.Location(sw.lat, ne.lon)
        
        TL = self.mm_provider.locationCoordinate(nw).zoomTo(zoom)
        
        # don't need this (note to self when trying to merge in to core...)
        # tiles = ModestMaps.TileQueue()
        
        cur_lon = sw.lon
        cur_lat = ne.lat        
        max_lon = ne.lon
        max_lat = sw.lat
        
        x_off = 0
        y_off = 0
        tile_x = 0
        tile_y = 0
        
        tileCoord = TL.copy()
        
        while cur_lon < max_lon :

            y_off = 0
            tile_y = 0
            
            while cur_lat > max_lat :

                # don't need this (note to self when trying to merge in to core...)
                # tiles.append(ModestMaps.TileRequest(self.mm_provider, tileCoord, ModestMaps.Core.Point(x_off, y_off)))
                
                y_off += self.mm_provider.tileHeight()
                
                tileCoord = tileCoord.down()
                loc = self.mm_provider.coordinateLocation(tileCoord)
                cur_lat = loc.lat

                tile_y += 1
                
            x_off += self.mm_provider.tileWidth()            
            cur_lat = ne.lat
            
            tile_x += 1
            tileCoord = TL.copy().right(tile_x)

            loc = self.mm_provider.coordinateLocation(tileCoord)
            cur_lon = loc.lon

        self.map_width = int(self.mm_provider.tileWidth() * tile_x)
        self.map_height = int(self.mm_provider.tileHeight() * tile_y)

        #
        # Work out the various dimensions
        # Adjust the map_width and map_height variables
        # as needed to ensure coverage for all
        # the sections and that rows have an
        # even number of columns
        #
        
        # flip!
        #
        # I can't remember why I thought this was a good idea
        # but it is the way it works now - something to do with
        # always generating the source map image in landscape
        # mode...I think

        self.width = int(self.paper_height * self.paper_dpi)
        self.height = int(self.paper_width * self.paper_dpi)
        self.margin = int(self.paper_margin * self.paper_dpi)

        self.sect_width = int((self.width - (self.margin * 4)) / 4)
        self.sect_height = int((self.height - (self.margin * 2)) / 2)

        mod_w = (self.map_width % self.sect_width)
        mod_h = (self.map_height % self.sect_height)

        #
        # Adjust adjust adjust
        #

        self.map_width += mod_w
        self.map_height += mod_h

        if ((self.map_width / self.sect_width) % 2) :
            self.map_width += self.sect_width

        count_w = (self.map_width / self.sect_width)
        count_h = (self.map_height / self.sect_height)

        self.map_width = count_w * self.sect_width
        self.map_height = count_h * self.sect_height

        # print "count : %s, %s" % (count_w, count_h)
        # print "dims : %s, %s" % (self.map_width, self.map_height)

        self.pages = (count_w * count_h) + 2
        self.sheets = int(math.ceil(float(self.pages) / float(8)))

        # print "map : %s, %s" % (self.map_width, self.map_height)
        # print "pages : %s" % self.pages
        # print "sheets : %s" % self.sheets
        # print "sect w:%s h:%s" % (self.sect_width, self.sect_height)

        self.sect_map = {
            'a' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 0},
            'b' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 0},
            
            'c' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : self.sect_width},
            'd' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : self.sect_width}, 
            
            'e' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 2 * self.sect_width},
            'f' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 2 * self.sect_width},
            
            'g' : {'rotate' : Image.ROTATE_270, 'x' : 0, 'y' : 3 * self.sect_width},
            'h' : {'rotate' : Image.ROTATE_90, 'x' : self.sect_height, 'y' : 3 * self.sect_width}
            }

        # profit!
        
    #
    #
    #
    
    def pil2cairo(self, im) :
        
        (w,h) = im.size
    
	mode = cairo.FORMAT_ARGB32

        data = im.tostring()
        a = array.array('B', data)

        return cairo.ImageSurface.create_for_data (a, mode, w, h, (w * 4))

    #
    #
    #
    
    def cairo2pil(self, surface) :

	mode='RGBA'
        
        width = surface.get_width()
        height = surface.get_height()
    
        return Image.frombuffer(mode, (width, height), surface.get_data(), "raw", mode, 0, 1)

    #
    #
    #
    
    def draw_page (self, pm, current_page) :

        piece_idx = current_page - 1
    
        if not self.pieces_map.has_key(piece_idx) :
            return

        page_idx = str(self.idx_for_page_number(current_page))
        
        sect = self.page_map[page_idx]
        meta = self.sect_map[sect]
    
        (row, col) = self.pieces_map[ piece_idx ]
    
        key = "%s-%s" % (row, col)

        img = self.pieces[row][col]
        img = self.draw_pg_number(img, current_page)
        img = img.transpose(meta['rotate'])
    
        x = int(meta['x'])
        y = int(meta['y'])

        x += self.margin
        y += self.margin

        if sect in ('e', 'f', 'g', 'h') :
            y += (self.margin * 2)
        
        pm.paste(img, (x,y))

    #
    #
    #
    
    def draw_pg_number (self, im, pg_number) :

        (w, h) = im.size

        if pg_number % 2 :
            # odd
            x = w - 30
        else :
            # even
            x = 30
        
        y = h - 30
        
        pm = self.pil2cairo(im)
    
        ctx = cairo.Context(pm)        
        ctx.move_to(x, y)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(55, 55, 55)
        ctx.fill()
        
        ctx = cairo.Context(pm)
        ctx.rectangle(x - 10, y - 10, 20, 20)
        ctx.set_source_rgb(0, 0, 0)
        ctx.set_line_width(2)
        ctx.stroke()
        
        ctx = cairo.Context(pm)        
        ctx.set_source_rgb(0, 0, 0)    
        
	offset_x, offset_y = self.offsets_for_page_number(ctx, pg_number)
                
        x = x + offset_x
        y = y - offset_y

        ctx.move_to(x, y)
        ctx.show_text(str(pg_number))

        self.draw_joiners(pm, pg_number)
        return self.cairo2pil(pm)

    #
    #
    #
    
    def draw_joiners (self, surface, pg_number) :
   
        piece = pg_number - 1
        (row, col) = self.pieces_map[ piece ]

        count_rows = len(self.pieces)
        count_cols = len(self.pieces[row])
    
        n = None
        s = None
        e = None
        w = None
        
        # north
    
        if row != 0 :
            n = ((row - 1), col)
        
        # south

        if row != (count_rows - 1) :
            s = ((row + 1), col)
        
        # west

        if col != 0 :
            if pg_number % 2 == 0:            
                w = (row, (col - 1))

        # east

        if col != (count_cols - 1) :
            if pg_number % 2 != 0:
                e = (row, (col + 1))

        #
    
        width = surface.get_width()
        height = surface.get_height()
    
        if n :
            x = int(width / 2)
            y = 30
            self.draw_joiner(surface, x, y, n)

        if s :
            x = int(width / 2)
            y = height - 30
            self.draw_joiner(surface, x, y, s)

        if e :
            x = width - 30
            y = int(height / 2)
            self.draw_joiner(surface, x, y, e)

        if w :
            x = 30
            y = int(height / 2)
            self.draw_joiner(surface, x, y, w)

    #
    #
    #
    
    def draw_joiner(self, surface, x, y, pos) :

        key = "%s-%s" % (pos[0], pos[1])
        pg_number = self.reverse_map[key] + 1

        radius = 10

        # please make this semi-transparent
        
        ctx = cairo.Context(surface)        
        ctx.move_to(x, y)
        ctx.arc(x, y, radius, 0, 360)
        ctx.set_source_rgba(255, 255, 255, 112)
        ctx.fill()
        
        ctx = cairo.Context(surface)
        ctx.arc(x, y, radius, 0, 360)
        ctx.set_source_rgb(0, 0, 0)
        ctx.set_line_width(2)
        ctx.stroke()
        
        ctx = cairo.Context(surface)        
        ctx.set_source_rgb(0, 0, 0)    

	offset_x, offset_y = self.offsets_for_page_number(ctx, pg_number)
                
        x = x + offset_x
        y = y - offset_y

        ctx.move_to(x, y)
        ctx.show_text(str(pg_number))

    #
    #
    #
    
    def offsets_for_page_number(self, ctx, pg_number) :

        # http://www.tortall.net/mu/wiki/CairoTutorial#tips-and-tricks
    
        fascent, fdescent, fheight, fxadvance, fyadvance = ctx.font_extents()
        
        xbearing, ybearing, width, height, xadvance, yadvance = (
            ctx.text_extents(str(pg_number)))
        
        x = 0.5 - xbearing - width / 2
        y = fdescent + fheight / 2

        # wtf do I need to adjust these again by hand?
        
    	return ((x - 1), (y - 10))
        
    #
    #
    #
    
    def idx_for_page_number(self, pg) :
    
        while pg > 8 :
            pg -= 8

        return pg


if __name__ == '__main__' :

    height = 11
    width = 8.5
    margin = .25
    dpi = 144

    bbox = (45.482882,-73.619899,45.532687,-73.547801)
    # bbox = (40.6977996826,-73.9747772217,40.7119102478,-73.9369964600)
    zoom = 16

    out = "example.pdf"
    
    pm = pocketMMap(height, width, margin, dpi)
    pm.load_provider('MICROSOFT_ROAD')
    pm.draw(bbox, zoom)
    pm.save(out)
